{Feedback Servo Driver}

CON
        _clkmode = xtal1 + pll16x    'Standard clock mode 
        _xinfreq = 5_000_000         '* crystal frequency = 80 MHz                    
      '  dutyScale = 360, dcmin = 33, dcrange = 961  'feedback scaling constants
        gain = 10                    'speed proportional to error
        
VAR                                       
  long  MyParm[10]                'passing addresses, etc to pasm cog
  byte  Cog
  long  PulseWidth                 'control pulse width in clock tics
  long  theta                      'feedback angle in degrees
  long  mode                       '0 => velocity mode, <>0 => position mode
  long  commanded                  'commanded angle               ' 
  long  OneUSec                    'clock tics in one microsecond
  
PUB start(CPin, FPin)                   'start the control cog
  OneUsec := clkfreq/1_000_000           'constant
  MyParm[0] := @PulseWidth               'address of pulse width to motor
  MyParm[1] := @Mode
  MyParm[2] := @theta                    'feedback: actual angle
  MyParm[3] := @Commanded                'commanded angle
  MyParm[4] := clkfreq / 200              'tics in 1/200th second (5 msec)               
  MyParm[5] := CPin                      'control pin number
  MyParm[6] := 1 << FPin                 'feedback pin mask 
  cog := cognew(@BCog, @MyParm)           'start up the pasm cog
  return cog            


PUB SetPos(degrees)
    Commanded := degrees               'save commanded angle
    Mode := 1                          'set position mode
    return theta                       'this is really not useful since it isnt done yet    

PUB GetPos
   return theta
    
PUB Width(Usec)                              'set pulse width
    mode := 0                                 'velocity mode
    pulsewidth := USec * OneUSec              'scale to clock tics
    
DAT   '***********pasm cog******************
BCog    org   0                'generates control pulse, returns DC in range 33..994               '
        mov AdPar,Par          'get the hub address of control PulseWidth
        rdlong AdPWidth, AdPar
        add AdPar, #4           'to address of mode
        rdlong AdMode, AdPar '
        add AdPar, #4           'to address of theta
        rdlong AdTheta, AdPar
        add AdPar, #4             'to address of commanded angle
        rdlong AdCommand, AdPar  
        add AdPar, #4             'to constant clk tics in 4 msec
        rdlong Hz200, AdPar
        add AdPar, #4             'to control pin number
        rdlong ContPin, AdPar
        add AdPar, #4              'to feedback pin mask
        rdlong FBPinMask, AdPar
        mov work1, #%00100          'pina follows phsa[31]
        shl work1, #26               'to control field
        mov ctra, work1             'start up counter a
        movs ctra, ContPin          'control pin
        mov frqa, #1                'count by one
        mov CPinMask, #1             'form mask with single "one"
        shl CPinMask, ContPin        'scale
        mov dira, CPinMask           'enable output
        or dira, RFBMask            'engineering
        or dira, P2
        or dira, P3
        mov timer, cnt              'initialize 250 hz timer
        add timer, Hz200                  
loop    rdlong work1, AdPWidth      'get control pulse width
        neg phsa, work1             'start a pulse
        mov Count4, #4              'will return four feedback angles
l1      or outa, RFBMask            'scope
        waitpne FBPinMask, FBPinMask 'wait for feedback low
        waitpeq FBPinMask, FBPinMask 'then high
        mov work1, cnt               'beginning of high feedback pulse
        waitpne FBPinMask, FBPinMask 'wait for end of high pulse
        mov HiPer, cnt
        waitpeq FBPinMask, FBPinMask 'now wait for end of low pulse
        mov LowPer, cnt
        sub LowPer, HiPer           'calculate low period
        sub HiPer, work1            'and high period
        mov y, HiPer                'calculate total period
        add y, LowPer               '
        shr y, #1                   'keep within 16 bits
        mov x, HiPer                'set up dividend
        shl x, #9                   'scale by 1024
        call #divide                'calculate dc
        andn x, Hi16                'discard remainder
        sub x, DCMin                'translate to zero
        mov y, x
        mov x, #359
        call #multiply
        mov x, y
        mov y, dcRange
        call #divide                'scale to degrees
        andn x, hi16                'drop remainder
        wrlong x, Adtheta           'return actual angle
        mov myTheta, x              'get my copy of actual
        rdlong work1, AdMode wz     'check mode
        if_z jmp #Done              'we are velocity mode
        rdlong MyCommand, AdCommand 'get the commanded angle
        mov error, MyTheta
        sub error, MyCommand wz, wc    'signed error
        if_Z jmp #PWGood              'they are the same:  dead nuts on
        abs error, error              'absolute value of error
        cmp error, #3 wz, wc          'accuracy
        if_C jmp #PWGood               'as close as can be
        or outa, P2                   'we will correct        
        cmp MyTheta, MyCommand wc     'error too big 
        if_C jmp #CCW                 'current is less than commanded
        or outa, P3                  'one way or other  
        wrlong MvCW, AdPWidth
        jmp #Done
CCW     wrlong MvCCW, AdPWidth
        jmp #Done          
PWGood  wrlong Hold, AdPWidth        'stop servo
Done    andn outa, RFBMask
        andn outa, P2
        andn outa, P3 
        waitcnt timer, Hz200         '5 msec period
        djnz Count4, #L1             'four times (total of 20 msec)
        jmp #loop                   'forever

' Divide x[31..0] by y[15..0] (y[16] must be 0)
' on exit, quotient is in x[15..0] and remainder is in x[31..16]
'
divide  shl  y,#15    'get divisor into y[30..15]
        mov  t,#16    'ready for 16 quotient bits
:loop   cmpsub x,y wc 'y =< x? Subtract it, quotient bit in c
        rcl    x,#1    'rotate c into quotient, shift dividend
        djnz   t,#:loop 'loop until done
divide_ret ret          'quotient in x[15..0], 'remainder in x[31..16]


' Multiply x[15..0] by y[15..0] (y[31..16] must be 0)
' on exit, product in y[31..0]
'
multiply shl   x,#16   'get multiplicand into x[31..16]
         mov   t,#16   'ready for 16 multiplier bits
         shr   y,#1 wc 'get initial multiplier bit into c
:loop if_c add y,x wc 'if c set, add multiplicand to product
         rcr   y,#1 wc 'put next multiplier in c, shift prod.
         djnz  t,#:loop 'loop until done
multiply_ret ret 'return with product in y[31..0]         
                    
zero        long  0
RFBMask     long  %0010         'I/O 1
Hi16        long  $FFFF_0000    'high word
DcMin       long  33
DcRange     long  961
Hold        long  1500*80 
MvCW        long  1550*80        'not proportional
MvCCW       long  1450*80
P2          long  1 << 2
P3          long  1 << 3 
Hz200       res                 'clock tics in 4 msec  
AdPar       res                 'parameter list                                                     
AdPWidth    res                 'address of control pulse width in hub
AdMode      res                 'address of mode (veleocity/angle)
Adtheta     res                 'address of duty cycle feedback
AdCommand   res                 'address of commanded angle
MyTheta     res                 'my copy of actual angle
MyCommand   res                 'my copy of commanded angle
Error       res                 'signed error   
HiPer       res                 'high period in clks
LowPer      res                 'low period
ContPin     res                 'control pin number
FBPinMask   res                 'feedback pin mask
timer       res                 'clock at end of next period
work1       res                 'generally useful
CPinMask    res                 'control pin mask
Count4      res                 'update position 5 times per 20 msec
x           res                 'pasm divide/ multiply variables
y           res
t           res          